// Copyright © 2015 Université Paris-Sud, Written by Lénaïc Bagnères, lenaic.bagneres@u-psud.fr

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.


#ifndef EXPRESSION_HPP
#define EXPRESSION_HPP

#include <type_traits>
#include <typeinfo>
#include <memory>

#include "remove_reference_wrapper.hpp"
#include "reference_wrapper_if_lvalue.hpp"


class lazy_expr_tag { };


// lazy binary and unary operation

template <class T0, class T1>
struct lazy_binary_operation
{
public:
	
	using tag = lazy_expr_tag;
	
private:
	
	T0 m_a;
	
	T1 m_b;
	
public:
	
	lazy_binary_operation(T0 const & a, T1 const & b) : m_a(a), m_b(b) { }
	
	typename remove_reference_wrapper<T0>::type const & a() const { return m_a; }
	
	typename remove_reference_wrapper<T1>::type const & b() const { return m_b; }
};

template <class T>
struct lazy_unary_operation
{
public:
	
	using tag = lazy_expr_tag;
	
private:
	
	T m_a;
	
public:
	
	lazy_unary_operation(T const & a) : m_a(a) { }
	
	typename remove_reference_wrapper<T>::type const & a() const { return m_a; }
};


// lazy +

template <class T0, class T1>
struct lazy_plus : lazy_binary_operation<T0, T1>
{
	using lazy_binary_operation<T0, T1>::lazy_binary_operation;
};

template <class T0, class T1>
lazy_plus<T0, T1> make_lazy_plus(T0 a, T1 b) { return lazy_plus<T0, T1>(a, b); }

// lazy +

template <class T0, class T1>
struct lazy_minus : lazy_binary_operation<T0, T1>
{
	using lazy_binary_operation<T0, T1>::lazy_binary_operation;
};

template <class T0, class T1>
lazy_minus<T0, T1> make_lazy_minus(T0 a, T1 b) { return lazy_minus<T0, T1>(a, b); }

// lazy *

template <class T0, class T1>
struct lazy_multiplies : lazy_binary_operation<T0, T1>
{
	using lazy_binary_operation<T0, T1>::lazy_binary_operation;
};

template <class T0, class T1>
lazy_multiplies<T0, T1> make_lazy_multiplies(T0 a, T1 b) { return lazy_multiplies<T0, T1>(a, b); }

// lazy /

template <class T0, class T1>
struct lazy_divides : lazy_binary_operation<T0, T1>
{
	using lazy_binary_operation<T0, T1>::lazy_binary_operation;
};

template <class T0, class T1>
lazy_divides<T0, T1> make_lazy_divides(T0 a, T1 b) { return lazy_divides<T0, T1>(a, b); }

// lazy -

template <class T>
struct lazy_negate : lazy_unary_operation<T>
{
	using lazy_unary_operation<T>::lazy_unary_operation;
};

template <class T>
lazy_negate<T> make_lazy_negate(T a) { return lazy_negate<T>(a); }


// +

template
<
	class T0, class T1,
	class T0_is_lazy_expr = typename std::enable_if<std::is_same<typename std::decay<T0>::type::tag, lazy_expr_tag>::value>::type,
	class T1_is_lazy_expr = typename std::enable_if<std::is_same<typename std::decay<T1>::type::tag, lazy_expr_tag>::value>::type
>
auto operator +(T0 && a, T1 && b)
{
	return
		make_lazy_plus
		(
			reference_wrapper_if_lvalue(std::forward<T0>(a)),
			reference_wrapper_if_lvalue(std::forward<T1>(b))
		);
}

// -

template
<
	class T0, class T1,
	class T0_is_lazy_expr = typename std::enable_if<std::is_same<typename std::decay<T0>::type::tag, lazy_expr_tag>::value>::type,
	class T1_is_lazy_expr = typename std::enable_if<std::is_same<typename std::decay<T1>::type::tag, lazy_expr_tag>::value>::type
>
auto operator -(T0 && a, T1 && b)
{
	return
		make_lazy_minus
		(
			reference_wrapper_if_lvalue(std::forward<T0>(a)),
			reference_wrapper_if_lvalue(std::forward<T1>(b))
		);
}

// *

template
<
	class T0, class T1,
	class T0_is_lazy_expr = typename std::enable_if<std::is_same<typename std::decay<T0>::type::tag, lazy_expr_tag>::value>::type,
	class T1_is_lazy_expr = typename std::enable_if<std::is_same<typename std::decay<T1>::type::tag, lazy_expr_tag>::value>::type
>
auto operator *(T0 && a, T1 && b)
{
	return
		make_lazy_multiplies
		(
			reference_wrapper_if_lvalue(std::forward<T0>(a)),
			reference_wrapper_if_lvalue(std::forward<T1>(b))
		);
}

// /

template
<
	class T0, class T1,
	class T0_is_lazy_expr = typename std::enable_if<std::is_same<typename std::decay<T0>::type::tag, lazy_expr_tag>::value>::type,
	class T1_is_lazy_expr = typename std::enable_if<std::is_same<typename std::decay<T1>::type::tag, lazy_expr_tag>::value>::type
>
auto operator /(T0 && a, T1 && b)
{
	return
		make_lazy_divides
		(
			reference_wrapper_if_lvalue(std::forward<T0>(a)),
			reference_wrapper_if_lvalue(std::forward<T1>(b))
		);
}

// -

template
<
	class T,
	class T_is_lazy_expr = typename std::enable_if<std::is_same<typename std::decay<T>::type::tag, lazy_expr_tag>::value>::type
>
auto operator -(T && a)
{
	return make_lazy_negate(reference_wrapper_if_lvalue(std::forward<T>(a)));
}


// display_expr

template <class T>
void display_expr(std::ostream & out, T const & expr)
{
	out << typeid(expr).name();
}

template <class T0, class T1>
void display_expr(std::ostream & out, lazy_plus<T0, T1> const & expr)
{
	// TODO
}

template <class T0, class T1>
void display_expr(std::ostream & out, lazy_minus<T0, T1> const & expr)
{
	// TODO
}

template <class T0, class T1>
void display_expr(std::ostream & out, lazy_multiplies<T0, T1> const & expr)
{
	// TODO
}

template <class T0, class T1>
void display_expr(std::ostream & out, lazy_divides<T0, T1> const & expr)
{
	// TODO
}

template <class T>
void display_expr(std::ostream & out, lazy_negate<T> const & expr)
{
	// TODO
}

#endif
