// Copyright © 2015 Université Paris-Sud, Written by Lénaïc Bagnères, lenaic.bagneres@u-psud.fr

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.


#ifndef VEC2_HPP
#define VEC2_HPP

#include <vector>

#include <hopp/memory/view_ptr.hpp>
#include <hopp/container/optional.hpp>

#include "tree.hpp"


// vec2<T>

template <class T>
class vec2 : public std::vector<T> // non-conforming inheritance
{
public:
	
	using std::vector<T>::vector;
};

template <class T>
std::ostream & operator <<(std::ostream & out, vec2<T> const & vec2)
{
	out << "{ ";
	
	for (auto const & e : vec2)
	{
		out << e;
		if (&e != &vec2.back()) { out << ","; }
		out << " ";
	}
	
	out << "}";
	
	return out;
}


// AST node (operation)

enum class operation_type { none, plus, minus, multiplies, divides, negate };

template <class T>
class operation_t
{
private:
	
	operation_type m_type;
	
	hopp::view_ptr<T const> p_data;
	
	hopp::optional<T const> m_data;
	
	operation_t(operation_type const type) :
		m_type(type), p_data(), m_data()
	{ }
	
	operation_t(operation_type const type, T && data) :
		m_type(type), p_data(), m_data(data)
	{ }
	
	operation_t(operation_type const type, T const & data) :
		m_type(type), p_data(data), m_data()
	{ }
	
public:
	
	operation_t(T && data) :
		operation_t(operation_type::none, std::forward<T>(data))
	{ }
	
	operation_t(T const & data) :
		operation_t(operation_type::none, data)
	{ }
	
	static operation_t<T> make_plus()
	{
		return operation_t(operation_type::plus);
	}
	
	static operation_t<T> make_minus()
	{
		return operation_t(operation_type::minus);
	}
	
	static operation_t<T> make_multiplies()
	{
		return operation_t(operation_type::multiplies);
	}
	
	static operation_t<T> make_divides()
	{
		return operation_t(operation_type::divides);
	}
	
	static operation_t<T> make_negate()
	{
		return operation_t(operation_type::negate);
	}
	
	operation_type type() const { return m_type; }
	
	hopp::view_ptr<T const> data() const
	{
		if (m_data) { return hopp::view_ptr<T const>(*m_data); }
		else { return p_data; }
	}
};


// AST

template <class T>
using ast_t = tree<operation_t<T>>;


// +

template <class T>
ast_t<vec2<T>> operator +(vec2<T> && a)
{
	return ast_t<vec2<T>>(operation_t<vec2<T>>(std::forward<vec2<T>>(a)));
}

template <class T>
ast_t<vec2<T>> operator +(vec2<T> const & a)
{
	return ast_t<vec2<T>>(operation_t<vec2<T>>(a));
}


// + - * / -

template <class T>
ast_t<T> operator +(ast_t<T> const & a, ast_t<T> const & b)
{
	return ast_t<T>(operation_t<T>::make_plus(), { a, b });
}

template <class T>
ast_t<T> operator -(ast_t<T> const & a, ast_t<T> const & b)
{
	return ast_t<T>(operation_t<T>::make_minus(), { a, b });
}

template <class T>
ast_t<T> operator *(ast_t<T> const & a, ast_t<T> const & b)
{
	return ast_t<T>(operation_t<T>::make_multiplies(), { a, b });
}

template <class T>
ast_t<T> operator /(ast_t<T> const & a, ast_t<T> const & b)
{
	return ast_t<T>(operation_t<T>::make_divides(), { a, b });
}

template <class T>
ast_t<T> operator -(ast_t<T> const & a)
{
	return ast_t<T>(operation_t<T>::make_negate(), { a });
}


// display_expr

template <class T>
void display_expr(std::ostream & out, ast_t<T> const & expr)
{
	// TODO
}


// get_size

template <class T>
size_t get_size(vec2<T> const & terminal)
{
	return 0; // TODO
}

template <class T>
size_t get_size(ast_t<vec2<T>> const & expr)
{
	return 0; // TODO
}


// compute_from_expr

template <class T>
auto compute_from_expr(ast_t<vec2<T>> const & expr, size_t const i)
{
	return 0.0; // TODO
}

template <class T>
vec2<T> compute_from_expr(ast_t<vec2<T>> const & expr)
{
	vec2<T> r(get_size(expr));
	
	//#pragma omp parallel for
	for (size_t i = 0; i < r.size(); ++i)
	{
		r[i] = compute_from_expr(expr, i);
	}
	
	return r;
}

#endif
