// Copyright © 2015 Université Paris-Sud, Written by Lénaïc Bagnères, lenaic.bagneres@u-psud.fr

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include <iostream>

#include <vec0.hpp>
#include <vec1.hpp>
#include <vec2.hpp>

#include <hopp/random.hpp>
#include <hopp/algo/sum.hpp>


void test_lvalue()
{
	std::cout << "test_lvalue =" << std::endl;
	
	// vec0
	
	vec0<double> vec0_a = { 1.0, 2.0, 3.0, 4.0, 5.0 };
	vec0<double> vec0_b = { 6.0, 7.0, 8.0, 9.0, 1.0 };
	
	auto vec0_r = vec0_a + vec0_b;
	
	std::cout << "vec0_r = " << vec0_r << std::endl;
	
	// vec1
	
	vec1<double> vec1_a = { 1.0, 2.0, 3.0, 4.0, 5.0 };
	vec1<double> vec1_b = { 6.0, 7.0, 8.0, 9.0, 1.0 };
	
	auto vec1_r = vec1_a + vec1_b;
	
	std::cout << "vec1_r = "; display_expr(std::cout, vec1_r); std::cout << std::endl;
	std::cout << "       = " << compute_from_expr(vec1_r) << std::endl;
	
	// vec2
	
	vec2<double> vec2_a = { 1.0, 2.0, 3.0, 4.0, 5.0 };
	vec2<double> vec2_b = { 6.0, 7.0, 8.0, 9.0, 1.0 };
	
	auto vec2_r = +vec2_a + +vec2_b;
	
	std::cout << "vec2_r = "; display_expr(std::cout, vec2_r); std::cout << std::endl;
	std::cout << "       = " << compute_from_expr(vec2_r) << std::endl;
}


void test_rvalue()
{
	std::cout << "test_rvalue =" << std::endl;
	
	// vec0
	
	auto vec0_r = vec0<double>(5, 2) + vec0<double>(5, 1);
	
	std::cout << "vec0_r = " << vec0_r << std::endl;
	
	// vec1
	
	auto vec1_r = vec1<double>(5, 2) + vec1<double>(5, 1);
	
	std::cout << "vec1_r = "; display_expr(std::cout, vec1_r); std::cout << std::endl;
	std::cout << "       = " << compute_from_expr(vec1_r) << std::endl;
	
	// vec2
	
	auto vec2_r = +vec2<double>(5, 2) + +vec2<double>(5, 1);
	
	std::cout << "vec2_r = "; display_expr(std::cout, vec2_r); std::cout << std::endl;
	std::cout << "       = " << compute_from_expr(vec2_r) << std::endl;
}


void test_lvalue_and_rvalue()
{
	std::cout << "test_lvalue_and_rvalue =" << std::endl;
	
	// vec0
	
	vec0<double> vec0_a = { 1.0, 2.0, 3.0, 4.0, 5.0 };
	vec0<double> vec0_b = { 6.0, 7.0, 8.0, 9.0, 1.0 };
	
	auto vec0_r = (vec0_a + vec0_b) + vec0<double>(5, 1);
	
	std::cout << "vec0_r = " << vec0_r << std::endl;
	
	// vec1
	
	vec1<double> vec1_a = { 1.0, 2.0, 3.0, 4.0, 5.0 };
	vec1<double> vec1_b = { 6.0, 7.0, 8.0, 9.0, 1.0 };
	
	auto vec1_r = (vec1_a + vec1_b) + vec1<double>(5, 1);
	
	std::cout << "vec1_r = "; display_expr(std::cout, vec1_r); std::cout << std::endl;
	std::cout << "       = " << compute_from_expr(vec1_r) << std::endl;
	
	// vec2
	
	vec2<double> vec2_a = { 1.0, 2.0, 3.0, 4.0, 5.0 };
	vec2<double> vec2_b = { 6.0, 7.0, 8.0, 9.0, 1.0 };
	
	auto vec2_r = (+vec2_a + +vec2_b) + +vec2<double>(5, 1);
	
	std::cout << "vec2_r = "; display_expr(std::cout, vec2_r); std::cout << std::endl;
	std::cout << "       = " << compute_from_expr(vec2_r) << std::endl;
}


void test_compute()
{
	std::cout << "test_compute =" << std::endl;
	
	std::vector<double> a(10 * 1000 * 1000);
	std::vector<double> b(a.size());
	
	for (size_t i = 0; i < a.size(); ++i)
	{
		a[i] = hopp::random::uniform(1.0, 9.0);
		b[i] = hopp::random::uniform(1.0, 9.0);
	}
	
	// vec0
	
	vec0<double> vec0_a(a.begin(), a.end());
	vec0<double> vec0_b(b.begin(), b.end());
	
	auto const vec0_begin = hopp::now::seconds();
	auto vec0_r = (((vec0_a + vec0_b) * -(vec0_a - vec0_b)) / -vec0_b) / (vec0_a * vec0_b);
	auto const vec0_end = hopp::now::seconds();
	
	std::cout << "hopp::sum(vec0_r) = " << hopp::sum(vec0_r) << " computed in " << vec0_end - vec0_begin << " seconds" << std::endl;
	
	// vec1
	
	vec1<double> vec1_a(a.begin(), a.end());
	vec1<double> vec1_b(b.begin(), b.end());
	
	auto const vec1_begin_expr = hopp::now::seconds();
	auto vec1_r_expr = (((vec1_a + vec1_b) * -(vec1_a - vec1_b)) / -vec1_b) / (vec1_a * vec1_b);
	auto const vec1_end_expr = hopp::now::seconds();
	
	auto const vec1_begin = hopp::now::seconds();
	auto vec1_r = compute_from_expr(vec1_r_expr);
	auto const vec1_end = hopp::now::seconds();
	
	std::cout << "hopp::sum(vec1_r) = " << hopp::sum(vec1_r)
	          << " computed in " << (vec1_end_expr - vec1_begin_expr) + (vec1_end - vec1_begin) << " seconds"
	          << " (" << vec1_end_expr - vec1_begin_expr << " + " << vec1_end - vec1_begin << ")" << std::endl;
	
	// vec2
	
	vec2<double> vec2_a(a.begin(), a.end());
	vec2<double> vec2_b(b.begin(), b.end());
	
	auto const vec2_begin_expr = hopp::now::seconds();
	auto vec2_r_expr = (((+vec2_a + +vec2_b) * -(+vec2_a - +vec2_b)) / -+vec2_b) / (+vec2_a * +vec2_b);
	auto const vec2_end_expr = hopp::now::seconds();
	
	auto const vec2_begin = hopp::now::seconds();
	auto vec2_r = compute_from_expr(vec2_r_expr);
	auto const vec2_end = hopp::now::seconds();
	
	std::cout << "hopp::sum(vec2_r) = " << hopp::sum(vec2_r)
	          << " computed in " << (vec2_end_expr - vec2_begin_expr) + (vec2_end - vec2_begin) << " seconds"
	          << " (" << vec2_end_expr - vec2_begin_expr << " + " << vec2_end - vec2_begin << ")" << std::endl;
}


void test_long_expr()
{
	std::cout << "test_long_expr =" << std::endl;
	
	std::vector<double> a(500);
	std::vector<double> b(a.size());
	
	for (size_t i = 0; i < a.size(); ++i)
	{
		a[i] = hopp::random::uniform(1.0, 9.0);
		b[i] = hopp::random::uniform(1.0, 9.0);
	}
	
	// vec0
	
	vec0<double> vec0_a(a.begin(), a.end());
	vec0<double> vec0_b(b.begin(), b.end());
	
	auto const vec0_begin = hopp::now::seconds();
	auto vec0_r = ((((vec0_a + vec0_b) * -(vec0_a - vec0_b)) / -vec0_b) / (vec0_a * vec0_b)) +
	              ((((vec0_a + vec0_b) * -(vec0_a - vec0_b)) / -vec0_b) / (vec0_a * vec0_b)) -
	              ((((vec0_a + vec0_b) * -(vec0_a - vec0_b)) / -vec0_b) / (vec0_a * vec0_b)) *
	              ((((vec0_a + vec0_b) * -(vec0_a - vec0_b)) / -vec0_b) / (vec0_a * vec0_b)) /
	              -vec0_b +
	              ((((vec0_a + vec0_b) * -(vec0_a - vec0_b)) / -vec0_b) / (vec0_a * vec0_b)) +
	              ((((vec0_a + vec0_b) * -(vec0_a - vec0_b)) / -vec0_b) / (vec0_a * vec0_b)) -
	              ((((vec0_a + vec0_b) * -(vec0_a - vec0_b)) / -vec0_b) / (vec0_a * vec0_b)) *
	              ((((vec0_a + vec0_b) * -(vec0_a - vec0_b)) / -vec0_b) / (vec0_a * vec0_b)) /
	              -vec0_b;
	auto const vec0_end = hopp::now::seconds();
	
	std::cout << "hopp::sum(vec0_r) = " << hopp::sum(vec0_r) << " computed in " << vec0_end - vec0_begin << " seconds" << std::endl;
	
	// vec1
	
	vec1<double> vec1_a(a.begin(), a.end());
	vec1<double> vec1_b(b.begin(), b.end());
	
	auto const vec1_begin_expr = hopp::now::seconds();
	auto vec1_r_expr = ((((vec1_a + vec1_b) * -(vec1_a - vec1_b)) / -vec1_b) / (vec1_a * vec1_b)) +
	                   ((((vec1_a + vec1_b) * -(vec1_a - vec1_b)) / -vec1_b) / (vec1_a * vec1_b)) -
	                   ((((vec1_a + vec1_b) * -(vec1_a - vec1_b)) / -vec1_b) / (vec1_a * vec1_b)) *
	                   ((((vec1_a + vec1_b) * -(vec1_a - vec1_b)) / -vec1_b) / (vec1_a * vec1_b)) /
	                   -vec1_b +
	                   ((((vec1_a + vec1_b) * -(vec1_a - vec1_b)) / -vec1_b) / (vec1_a * vec1_b)) +
	                   ((((vec1_a + vec1_b) * -(vec1_a - vec1_b)) / -vec1_b) / (vec1_a * vec1_b)) -
	                   ((((vec1_a + vec1_b) * -(vec1_a - vec1_b)) / -vec1_b) / (vec1_a * vec1_b)) *
	                   ((((vec1_a + vec1_b) * -(vec1_a - vec1_b)) / -vec1_b) / (vec1_a * vec1_b)) /
	                   -vec1_b;
	auto const vec1_end_expr = hopp::now::seconds();
	
	auto const vec1_begin = hopp::now::seconds();
	auto vec1_r = compute_from_expr(vec1_r_expr);
	auto const vec1_end = hopp::now::seconds();
	
	std::cout << "hopp::sum(vec1_r) = " << hopp::sum(vec1_r)
	          << " computed in " << (vec1_end_expr - vec1_begin_expr) + (vec1_end - vec1_begin) << " seconds"
	          << " (" << vec1_end_expr - vec1_begin_expr << " + " << vec1_end - vec1_begin << ")" << std::endl;
	
	// vec2
	
	vec2<double> vec2_a(a.begin(), a.end());
	vec2<double> vec2_b(b.begin(), b.end());
	
	auto const vec2_begin_expr = hopp::now::seconds();
	auto vec2_r_expr = (((+vec2_a + +vec2_b) * -(+vec2_a - +vec2_b)) / -+vec2_b) / (+vec2_a * +vec2_b) +
	                   (((+vec2_a + +vec2_b) * -(+vec2_a - +vec2_b)) / -+vec2_b) / (+vec2_a * +vec2_b) -
	                   (((+vec2_a + +vec2_b) * -(+vec2_a - +vec2_b)) / -+vec2_b) / (+vec2_a * +vec2_b) *
	                   (((+vec2_a + +vec2_b) * -(+vec2_a - +vec2_b)) / -+vec2_b) / (+vec2_a * +vec2_b) /
	                   -+vec2_b +
	                   (((+vec2_a + +vec2_b) * -(+vec2_a - +vec2_b)) / -+vec2_b) / (+vec2_a * +vec2_b) +
	                   (((+vec2_a + +vec2_b) * -(+vec2_a - +vec2_b)) / -+vec2_b) / (+vec2_a * +vec2_b) -
	                   (((+vec2_a + +vec2_b) * -(+vec2_a - +vec2_b)) / -+vec2_b) / (+vec2_a * +vec2_b) *
	                   (((+vec2_a + +vec2_b) * -(+vec2_a - +vec2_b)) / -+vec2_b) / (+vec2_a * +vec2_b) /
	                   -+vec2_b;
	auto const vec2_end_expr = hopp::now::seconds();
	
	auto const vec2_begin = hopp::now::seconds();
	auto vec2_r = compute_from_expr(vec2_r_expr);
	auto const vec2_end = hopp::now::seconds();
	
	std::cout << "hopp::sum(vec2_r) = " << hopp::sum(vec2_r)
	          << " computed in " << (vec2_end_expr - vec2_begin_expr) + (vec2_end - vec2_begin) << " seconds"
	          << " (" << vec2_end_expr - vec2_begin_expr << " + " << vec2_end - vec2_begin << ")" << std::endl;
}


int main()
{
	std::cout << std::endl;
	
	int nb_test = 0;
	
	
	test_lvalue();
	std::cout << std::endl;
	
	
	test_rvalue();
	std::cout << std::endl;
	
	
	test_lvalue_and_rvalue();
	std::cout << std::endl;
	
	
	test_compute();
	std::cout << std::endl;
	
	
	test_long_expr();
	std::cout << std::endl;
	
	
	if (nb_test != 0)
	{
		std::cout << "expression = " << nb_test << " tests fail!" << std::endl;
		std::cout << std::endl;
	}
	
	return nb_test;
}
